"""
Tools to read and write audio files from many places in many different formats.
"""
from hear.stores import *
from hear.tools import AudioSegments, wf_func_to_wfsr_func
from hum.sound import Sound
